#!/usr/bin/perl 

use strict;
use CGI;
use File::Basename;
use Getopt::Long;
use HTML::Parser;
use LWP::UserAgent;
use Unicode::String qw(utf8 latin1);
use URI::URL;

use vars qw($program $version);
$program = 'Html2Wml';
$version = '0.3.4';

=head1 NAME

Html2Wml - Program that can convert HTML pages to WML pages

=head1 SYNOPSIS

in a shell:

    html2wml.cgi [--ascii] [--hreftmpl=TEMPLATE] 
                 [--nocollapse] [--nolinearize] 
                 [--debug] [--xmlcheck] <file|url>

as a CGI:

    /cgi-bin/html2wml.cgi?url=<url>


=head1 DESCRIPTION

Html2Wml converts HTML pages to WML pages, suitable for being viewed on 
a Wap device. The conversion can be done either on the command line to 
create static WML pages or on-the-fly by calling this program as a CGI. 

As of version 0.3, the resulting WML should be well-formed, and in most
cases valid. This is not guarantied but it should work for most HTML pages. 
To be more precise, the validity of the WML depends on the quality of the 
input HTML. Pages created with softwares that conform to W3C standard 
are most likely to produce valid WML. To check your HTML pages, your can 
use W3C's excellent software I<HTML Tidy>, written by Dave Raggett. 


=head1 OPTIONS

=over 4

=item ascii

When this option is on, named HTML entities are converted to US-ASCII 
using the same 7bit approximations as Lynx. 
By default, this is off, so that named entities are converted to Unicode 
characters. In both cases, numeric entities are kept untouched. 

=item collapse

This option tells Html2Wml to collapse redundant white space characters 
and empty paragraphs. This option is on by default, but you can desactivate 
this by using --nocollapse. 

This behavior is not really standard, but the aim is reducing the size of 
the output. WML pages are primarily intented for Wap devices, which usually 
have slow connections. The smaller the WML result is, the faster it can be 
downloaded. Furthermore, collapsing white spaces is the normal behavior 
for HTML pages. 

Empty paragraphs are also collapsed (this is really not standard), but it 
should avoid empty screens: the display of a Wap device is usually small, 
and it can be annoying to scroll down a lot because of many empty lines. 

=item debug

This option activates the debug mode. This prints the output result in HTML 
with line numbering and with the result of the XML check. This option is very 
useful for debugging as you can use any web browser for that purpose. 

=item hreftmpl

The link template is an internal engine that can be used to reconstruct
the links of a page in a customizable way. 

See the I<Links Reconstruction> section for more information. 

=item linearize

This options is on by default. It makes Html2Wml flattens the tables 
I<E<agrave> la> Lynx. I think it is better than trying to use WML tables 
because, contrary to HTML tables, they have extremely limited features (in
particular, they can't be nested). Therefore it's quite difficult to decide
what to do when you have three nested tables. Furthermore, calculations on
tables are quite CPU consuming, and Wap devices are not supposed to be
powerful. 

=item xmlcheck

When this option is on, it send the WML output to XML::Parser to check 
its well-formedness. 

=back


=head1 FEATURES

=head2 Actions

Actions are a feature similar to the SSI (Server Side Includes) available on
web servers like Apache. In order not to interfere with real SSI, but to keep 
their syntax easy to learn, it differs in very few points. 

=head3 Syntax

The syntax to execute an action is:

    <!-- [action param1="value" param2='value'] -->

Note that the angle brackets are part of the syntax. 
Except for that point, Actions syntax is very similar to SSI syntax. 

=head3 Available actions

=over 4

=item include 

B<Description>

Includes a file in the document at the current point. 
Please note that Html2Wml doesn't check nor parse the file, and if the file
cannot be found, will silently die (this is the same behavior as SSI). 

B<Parameters>

virtual=I<url>

=over 4

The file is get by http. 

=back

file=I<path>

=over 4

The file is read from the local disk. 

=back

B<Note>

If you use the C<file> parameter, an absolute path is recommend. 

=item fsize

B<Description>

Returns the size of a file at the current point of the document. 

B<Parameters>

You can use the same parameters as for the C<include> action. 

=back

=head3 Examples

To include a small navigation bar: 

    <!-- [include virtual="nav.wml"] -->


=head2 Links Reconstruction

This engine can be used when calling Html2Wml from command line. It allows 
you to reconstruct the links of a page so that they point to useful URI. 
The result URL can be customized by using the link template (this is the 
parameter of the C<hreftmpl> option). 

Please note this engine is currently very limited. Additional features will 
be added if requested. 

=head3 Syntax

The template is a string that contains the new URL. 
You can interpolate parameters by simply including them in the template 
between curly brackets: C<{I<param>}>

=head3 Available parameters

=over 4

=item C<URL>

This parameter contains the original URL from the C<href> or C<src> attribute. 

=item C<FILENAME>

This parameter contains the base name of the file. 

=item C<FILEPATH>

This parameter contains the leading path of the file. 

=item C<FILETYPE>

This parameter contains the suffix of the file. 

=back

=head3 Examples

To add a path option: 

    {URL}$wap

Using Apache, you can then add a Rewrite directive so that URL ending with 
C<$wap> will be redirected to Html2Wml. 

To change the base name of the file: 

    {FILEPATH}{FILENAME}_wap{FILETYPE}


=head1 CAVEATS

Currently, only the well-formedness of the resulting WML can be tested, not  
its validity. 

Inverted tags (like "<b>bold <i>italic</b></i>") may produce unexpected 
results. But only bad softwares do bad stuff like this. 


=head1 TODO

=over 4

=item *

add support for images (via external conversion)

=item *

add an engine to allow attributes conversion

=item *

add a support to convert forms

=item *

add a support to create binary WML, either by using an external compiler 
as wmlc from WML Tools, or by implementing this internally

=item *

add a validity check (for debugging purpose)

=back


=head1 VERSION HISTORY

=over 4

=item v0.3.4, Wednesday, December 20, 2000

corrected a bug in the de-nesting process

added features to the links reconstruction engine

corrected the local file disclosure vulnerability

until now, captions tags were always skipped; now they appears when tables
are linearized

=item v0.3.3, Monday, December 18, 2000

named entities are converted to real Unicode characters

non-breaking spaces (&nbsp;) are converted to spaces

=item v0.3.3 beta 4, Friday, December 15, 2000

links reconstruction is finalized

=item v0.3.3 beta 3, Friday, December 15, 2000

added the C<ascii> option to convert entities to US-ASCII

=item v0.3.3 beta 2, Friday, December 15, 2000

named entities conversion is finalized

=item v0.3.3 beta 1, Friday, December 14, 2000

added preliminary support for named entities conversion and images support

=item v0.3.2, Thursday, December 13, 2000

oops! file download was broken in 0.3.1; this is corrected now

ampersand are now coded as &amp; in links

a long standing bug that appeared in some cirumstances has been fixed

=item v0.3.1, Tuesday, December 12, 2000

added a small SSI-like engine (not heavily tested)

corrected a stupid bug: I forgot to add the line that translate the internal 
links so that the urls point to the CGI

modification of the user agent signature so that it includes the signature 
of the original client

rendering of header tags H1, H2 and H3 has been improved

added a links reconstruction engine (currently unfinished)

improved the debug mode when launched from shell

=item v0.3.0, Sunday, December 10, 2000

the syntax repair engine has been improved and should now be able to 
correct most usual HTML pages (I hope)

=item v0.3.0 beta 6, Saturday, December 9, 2000

corrected a stupid bug in the CGI mode

added the C<debug> option

corrected some mistakes in the documentation

added the LINKS section in the documentation

added several improvements and corrected several bugs

=item v0.3.0 beta 5, Thursday, December 6, 2000

added the C<collapse> option to control whether redundant white space 
chars are kept or not

added the documentation for the options

=item v0.3.0 beta 4, Monday, December 4, 2000

added the C<linearize> option (on by default) so that it linearizes 
tables

normal texts (i.e. not inside <pre> tags) are cleaned by stripping 
multiples spaces and lines

the syntax repair engine has been rewritten, but it does not work in 
all cases

=item v0.3.0 beta 3, Sunday, November 19, 2000

first try to add syntax check/repair

=item v0.2.3, Saturday, November 18, 2000

added URI translation

=item v0.2.2, Thursday, November 15, 2000

added XML checker; the output is given to Expat to check that it is 
well-formed XML. Note that XML::Parser is loaded on demand.

added options support (using Getopt::Long)

=item v0.2.1, Tuesday, November 14, 20000

cleaned up a few things;

added some attributes to the allow list

added url resolution

=item v0.2.0, Saturday, November 11, 2000

first functional version; now uses HTML::Parser v3

=item v0.1.0, Friday, November 10, 2000

prototype version

=back


=head1 LINKS

=over 2

=item Html2Wml -- HTML to WML converter

http://www.resus.univ-mrs.fr/~madingue/techie/html2wml.html

=item HTML Tidy

http://www.w3.org/People/Raggett/tidy

=item WML Tools

http://pwot.co.uk/wml/

=item wApua -- Wap Wml browser written in Perl/Tk

http://fsinfo.cs.uni-sb.de/~abe/wApua/

=item Tofoa -- Wap emulator written in Python

http://tofoa.free-system.com/

=item WML Browser -- Free WML browser for Linux

http://www.wmlbrowser.org/

=back


=head1 ACKNOWLEDGEMENTS

Werner Heuser - for his numerous ideas, advices and his help for the debugging


=head1 AUTHOR

SE<eacute>bastien Aperghis-Tramoni <madingue@resus.univ-mrs.fr>


=head1 COPYRIGHT

Html2Wml is Copyright (C)2000 SE<eacute>bastien Aperghis-Tramoni.

This program is free software. You can redistribute it and/or modify it 
under the terms of either the Perl Artistic License or the GNU General 
Public License, version 2 or later. 

=cut


# 
# globals
# 
use vars qw($cgi);
$cgi = 0;
my $result;
my %options = (
    help       => 0,    ## show the usage and exit
    version    => 0,    ## show the program name and version and exit
    
    ascii      => 0,    ## convert named entities to US-ASCII
    collapse   => 1,    ## collapse white space characters
    debug      => 0,    ## activate the debug mode
    hreftmpl   => '',   ## contain the template for xlate_url()
    linearize  => 1,    ## suppress the tables tags
    xmlcheck   => 0,    ## perform a well-formedness check (using XML::Parser)
);
my $xmlckres = '';

## used by the html parser
use vars qw(%state);
%state = (
    doc_uri  => '',  ## document absolute URI
    output   => '',  ## buffer for storing output
    cardsize => 0,   ## size of the card
    skip     => 0,   ## skip switch (on/off)
    stack    => []   ## tag stack
);

my %entities;  ## named entities conversion table

# 
# The following two hashes are based on the WML DTD. They are the hardcoded 
# conversion tables. They also describe the legal syntax of WML tags. 
# 
my %dtdent = ();
    $dtdent{emph}   = 'em,strong,b,i,u,big,small';
    $dtdent{layout} = 'br';
    $dtdent{text}   = $dtdent{emph};
    $dtdent{flow}   = "$dtdent{text},$dtdent{layout},img,anchor,a,table";
    $dtdent{fields} = "$dtdent{flow},input,select,fieldset";

my %with = (
    html    => { action => 'replace',  new_value => 'wml'  }, 
    wml     => { action => 'keep',     nest => 'head,card'  }, 
    
    ## header tags
   # head    => { action => 'keep',     nest => 'meta' }, 
   # meta    => { action => 'keep',     nest => 'EMPTY' }, 
    title   => { action => 'skip' }, 
    style   => { action => 'skip' }, 
    script  => { action => 'skip' }, 
    
    ## structural tags
    body    => { action => 'replace',  new_value => 'card' }, 
    card    => { action => 'keep',     nest => 'do,p,pre' }, 
    h1      => { action => 'replace',  new_value => 'p',  special => 'big,b' }, 
    h2      => { action => 'replace',  new_value => 'p',  special => 'big' }, 
    h3      => { action => 'replace',  new_value => 'p',  special => 'b' }, 
    h4      => { action => 'replace',  new_value => 'p' }, 
    h5      => { action => 'replace',  new_value => 'p' }, 
    h6      => { action => 'replace',  new_value => 'p' }, 
    li      => { action => 'replace',  new_value => 'p' }, 
    dt      => { action => 'replace',  new_value => 'p' }, 
    dd      => { action => 'replace',  new_value => 'p' }, 
    div     => { action => 'replace',  new_value => 'p' }, 
    p       => { action => 'keep',     nest => "$dtdent{fields},do" }, 
    br      => { action => 'keep',     nest => 'EMPTY' }, 
    pre     => { action => 'keep',     nest => 'a,br,i,b,em,strong,input,select' }, 
    tt      => { action => 'replace',  new_value => 'pre' }, 
    
    ## tables tags
    table   => { action => 'keep',     nest => 'tr',  attributes => 'colspan,rowspan' }, 
    caption => { action => 'skip' }, 
   'tr'     => { action => 'keep',     nest => 'td' }, 
    th      => { action => 'replace',  new_value => 'td' }, 
    td      => { action => 'keep',     nest => "$dtdent{emph},$dtdent{layout},img,a,anchor" }, 
    
    ## link tags
    a       => { action => 'keep',     nest => 'br,img',  attributes => 'href,title' }, 
    img     => { action => 'keep',     nest => 'EMPTY',  attributes => 'src,alt,align' }, 
    
    ## form tags -- currently not handled
   # select  => { action => 'keep',     nest => 'optgroup,option' }, 
   # input   => { action => 'keep',     nest => 'EMPTY' }, 
    
    ## font tags
    em      => { action => 'keep',     nest => $dtdent{flow} }, 
    strong  => { action => 'keep',     nest => $dtdent{flow} }, 
    b       => { action => 'keep',     nest => $dtdent{flow} }, 
    i       => { action => 'keep',     nest => $dtdent{flow} }, 
    u       => { action => 'keep',     nest => $dtdent{flow} }, 
    big     => { action => 'keep',     nest => $dtdent{flow} }, 
    small   => { action => 'keep',     nest => $dtdent{flow} }
);


# 
# main
# 
$| = 1;

fileparse_set_fstype('Unix');  ## this is because I use fileparse() to 
                               ## split the URL fragments

## CGI security options
$CGI::POST_MAX = 1024 * 1;  # max 1K posts
$CGI::DISABLE_UPLOADS = 1;  # no uploads

load_entities();

if(@ARGV) {
    ## launched from shell
    
    ## getting options
    GetOptions(\%options, qw(help version  
                             ascii! collapse! debug! hreftmpl:s linearize! xmlcheck!));
    version() if $options{version};
    usage() if $options{help};
    usage() unless @ARGV;
    apply_options();
    
    ## converting the file
    $result = html2wml(shift);
    
} else {
    ## launched from web
    $cgi = new CGI;
    
    ## get the options
    for my $option (keys %options) {
        $options{$option} = $cgi->param($option) if defined $cgi->param($option)
    }
    
    apply_options();
    
    print $cgi->header if $options{debug};
    
    $result = html2wml($cgi->param('url')) or cgiusage();
}


## XML check
if($options{xmlcheck}) {
    eval {
      require XML::Parser;
      my $xmlparser = new XML::Parser Style => 'Tree', ErrorContext => 2;
      $xmlparser->parse($result);
    };
    $xmlckres = $@ ? $@ : "Well-formed";
}

if($options{debug}) {
    my $i = 1;
    $result =~ s/^/@{[sprintf '%3d', $i++]}: /gm;  ## add lines number
    
    if($cgi) {
        print qq|<html>\n<head>\n<title>$program -- Debug Mode</title>\n</head>\n|, 
              qq|<body>\n<h1>$program -- Debug Mode</h1>\n|, 
              qq|<p>This is the result of the conversion of the document |, 
              qq|<a href="$state{doc_uri}">$state{doc_uri}</a></p>\n|, 
              qq|<hr />\n|, 
              htmlize($result), 
              qq|<hr />\n<p>Result of XML check:</p>\n|, 
              htmlize($xmlckres), 
              qq|</body>\n</html>|
    } else {
        my $s = "$program -- Debug Mode\n";
        print $s, '-'x length($s), "\n", 
              $result, "\n", ' -'x 5, "\n", 
              $xmlckres, "\n"
    }
    
} else {
    print $cgi->header(-type => 'text/vnd.wap.wml') if $cgi;
    print $result;
}



# 
# apply_options()
# -------------
sub apply_options {
    if($options{linearize}) {
        delete @with{qw(table tr td th)};
        $with{'caption'} = { action => 'replace', new_value => 'p', special => 'b'};
        $with{'tr'} = { action => 'replace', new_value => 'p' };
        $with{'td'} = { action => 'replace', new_value => 'p' };
    }
    
    if($options{debug}) {
        $options{xmlcheck} = 1;
    }
}


# 
# html2wml()
# --------
sub html2wml {
    my $url = shift;
    my $file;
    my $type;
    my %attr;
    my $converter = new HTML::Parser api_version => 3;
    my $date = localtime;
    
    ## read the file 
    if($url =~ m{http://}) {  ## absolute uri
        ($file,$type) = get_url($url)
    
    } elsif(not $cgi and -f $url) {  ## local file
        $file = read_file($url)
    
    } else {  ## absolute url relative to the server
        ($file,$type) = get_url( URI::URL->new($url, $cgi->url)->abs )
    }
    
    ## if it's an image, call send_image()
    if($url =~ /\.(?:gif|jpg|png)$/i or $type =~ /image/) {
        @_ = ($file, $url);
        goto &send_image
    }
    
    ## gather information from the HTML page
    if($file =~ m/<title>(.*)<\/title>/is) {
        $attr{title} = clean_spaces($1)
    }
    
    ## conversion
    $state{skip} = 0;
    $state{output} = q|<?xml version="1.0"?>| . "\n" 
                   . q|<!DOCTYPE wml PUBLIC "-//WAPFORUM//DTD WML 1.2//EN" |
                   . q|"http://www.wapforum.org/DTD/wml12.dtd">| . "\n" 
                   .qq|<!-- Converted by $program $version on $date -->\n|;
    
    $converter->unbroken_text(1);
    $converter->handler(start       => \&start,   'tagname, attr, attrseq, text');
    $converter->handler(end         => \&end,     'tagname, text');
    $converter->handler(text        => \&text,    'text, is_cdata');
    $converter->handler(comment     => \&comment, 'tokens');
    $converter->handler(declaration => '');
    $converter->handler(process     => '');
    
    $converter->parse($file);
    $converter->eof;
    
    ## flush the stack
    while(my $tag = pop @{$state{stack}}) {
        $state{output} .= "</$tag>"
    }
    
    ## convert the named entities to Unicode characters
    my $unicode = latin1( convert_entities($state{output}) );
    $state{output} = $unicode->utf8;
    
    ## post-conversion clean-up
    $state{output} =~ s/\015\012|\012|\015/\n/go;  ## converts CR/LF to native eol
    
    $state{output} =~ s|\s+>|>|go;
    $state{output} =~ s|\s+/>|/>|go;
    $state{output} =~ s|<p>\s*</p>||go  if $options{collapse};
    $state{output} =~ s|<p>\s*<br/>\s*</p>||go  if $options{collapse};
    $state{output} =~ s|<p>\s*&nbsp;\s*</p>||go if $options{collapse};
    $state{output} =~ s/ +/ /go         if $options{collapse};
    $state{output} =~ s/\t+/ /go        if $options{collapse};
    $state{output} =~ s/\n+/\n/go       if $options{collapse};
    $state{output} =~ s/(?: \n)+/\n/go  if $options{collapse};
    
    ## set the title of the card
    if(length $attr{title}) {
        $state{output} =~ s/<card/<card title="$attr{title}"/;
    }
    
    return $state{output}
}


# 
# get_url()
# -------
# This function gets and returns the file from the given URI. 
# 
sub get_url {
    my $uri = shift;
    my $quiet = shift || 0;
    my $ua = new LWP::UserAgent;
    $ua->agent( $cgi ? $cgi->user_agent." [through $program/$version ".$ua->agent."]" 
                     : "$program/$version ".$ua->agent );
    my $request = new HTTP::Request GET => $uri;
    my $response = $ua->request($request);
    
    if($response->is_error) {
        return $quiet ? '' : $response->error_as_HTML
    }
    
    $state{doc_uri} = $uri;
    return wantarray ? ($response->content, $response->content_type) : $response->content
}


# 
# read_file()
# ---------
# This function reads and returns the file from the local disk. 
# 
sub read_file {
    my $file = shift;
    my $quiet = shift || 0;
    open(FILE, $file) or ($quiet ? return '' : die "can't read file '$file': $! ");
    local $/ = undef;
    $file = <FILE>;
    close(FILE);
    return $file
}


# 
# send_image()
# ----------
# This function allow Html2Wml to send WBMP images to the client. 
# Currently, it send an empty hardcoded image, but support for 
# conversion from common formats (GIF, JPEG, PNG) will be added soon. 
# 
sub send_image {
    my $data = shift;
    my $path = shift;
    
    my $pixel = pack 'C*', 0, 0, 1, 1, 0xFF;  ## this is one white pixel
    
    ## [here there will be the conversion]
    
    print $cgi->header(-type => 'image/wbmp'), $pixel;
    exit
}


# 
# convert_entities()
# ----------------
# This function converts the named entities from HTML to Unicode 
# characters (with the exception of the entities known to WML of course). 
# 
sub convert_entities {
    my $text = shift;
    my $ascii = $options{ascii};
    
    my $code = q|  while($text =~ /&(\w+);/g) {               |
             . q|      my $ent = $1;                          |
             . q|      if(exists $entities{$ent}) {           |
    .($ascii ? q|          my $chr = $entities{$ent}[1];      |
             : q|          my $chr = chr($entities{$ent}[0]); | )
             . q|          $text =~ s/&$ent;/$chr/g           | 
             . q|      }                                      |
             . q|  }                                          |;
    
    eval $code;
    
    return $text
}


# 
# clean_spaces()
# ------------
sub clean_spaces {
    my $str = shift;
    $str =~ s/ +/ /go;
    $str =~ s/^ //;
    return $str
}


# 
# Functions used by the HTML parser
#  - start() is called on a start tag event
#  - end() is called on an end tag event
#  - text() is called on a text event
# 
  
sub start {
    my($tag, $attr) = @_;
    my $curr_tag = ($with{$tag}{action} eq 'replace' ? $with{$tag}{new_value} : $tag);
    
    ## reconstruct well-formed attributes list with only the allowed ones
    $attr = join ' ', map { $_.'="' . (/href|src/ ? xlate_url($attr->{$_}) : $attr->{$_}) . '"' if $attr->{$_} } 
                      split(',', $with{$curr_tag}{attributes});
    $attr = ' ' . $attr if length $attr;
    
    ## set the skip mode state
    $state{skip} = 1 if $with{$curr_tag}{action} eq 'skip';
    
   # print "(start tag) $tag => action: ", 
   #       ($with{$tag}{action} ? $with{$tag}{action} : 'clear'), 
   #       ($curr_tag ne $tag ? " with $curr_tag " : ''), 
   #       ($attr? " attributes:$attr" : ''), 
   #       "<br />\n" if $options{debug};
    
    ## syntax check/repair
    if(@{$state{stack}} and $with{$curr_tag}{action} eq 'keep' and $with{$curr_tag}{nest} ne 'EMPTY') {
        while(my $prev_tag = pop @{$state{stack}}) {
            if($with{$prev_tag}{nest} =~ /\b$curr_tag\b/) {
                push @{$state{stack}}, $prev_tag;
                last
            }
            $state{output} .= "</$prev_tag>";
        }
    }
    
    ## simple tag translation
    if($with{$curr_tag}{action} eq 'keep') {
        if($with{$curr_tag}{nest} eq 'EMPTY') {
            $state{output} .= "<$curr_tag$attr/>"
        } else {
            $state{output} .= "<$curr_tag$attr>";
            push @{$state{stack}}, $curr_tag;
        }
    
    } else {
        ## do nothing
    }
    
    ## additional rendering effects
    if(defined $with{$tag}{special}) {  ## note that it is $tag, not $curr_tag
        for my $t (split ',', $with{$tag}{special}) {
            $state{output} .= "<$t>"
        }
    }
}
  
sub end {
    my($tag) = @_;
    my $curr_tag = ($with{$tag}{action} eq 'replace' ? $with{$tag}{new_value} : $tag);
    
   # print "( end tag ) $curr_tag, stack = (@{$state{stack}}) <br />\n" if $options{debug};
    
    $state{skip} = 0 if $with{$tag}{action} eq 'skip';
    return if $with{$tag}{nest} eq 'EMPTY';
    
    ## additional rendering effects
    if(defined $with{$tag}{special}) {  ## note that it is $tag, not $curr_tag
        for my $t (reverse split ',', $with{$tag}{special}) {
            $state{output} .= "</$t>"
        }
    }
    
    if(${$state{stack}}[-1] eq $curr_tag  and  $with{$curr_tag}{action} eq 'keep') {
        $state{output} .= "</$curr_tag> ";
        pop @{$state{stack}};
        
        ## add a para tag if we're on the card node
        if(${$state{stack}}[-1] eq 'card') {
            $state{output} .= "\n<p>";
            push @{$state{stack}}, 'p';
        }
    
    } else {
        ## do nothing
    }
    
    ## check current card size
    if($state{cardsize} ) {
        ## todo ...
    }
}
  
sub text {
    my($text) = @_;
    $text = clean_spaces($text) if $options{collapse} and ${$state{stack}}[-1] ne 'pre';
    $state{cardsize} += length $text;
    $state{output} .= $text unless $state{skip}
}

sub comment {
    my($comment) = @_;
    local $_;
    
    $comment = join '', @$comment;
    
    ## SSI engine
    if($comment =~ /^\s*\[(\w+)\s+(.*)\]\s*$/) {
        my $element = $1;
        my %attributes = map { /\G(\w+)=["']([^"']+)["']/g } split /\s+/, $2;
        
        for my $attr (keys %attributes) {
            if($attr eq 'virtual' and $attributes{virtual} !~ /^http:/) {
                $attributes{virtual} = URI::URL->new( $attributes{virtual}, $state{doc_uri} )->abs
            }
        }
        
        for($element) {
            /include/ and do {
                if(defined $attributes{virtual}) { $state{output} .= get_url($attributes{virtual}, 1) }
                elsif(defined $attributes{file}) { $state{output} .= read_file($attributes{file}, 1) }
            };
            
            /fsize/ and do {
                if(defined $attributes{virtual}) { $state{output} .= length get_url($attributes{virtual}, 1) }
                elsif(defined $attributes{file}) { $state{output} .= length read_file($attributes{file}, 1) }
            };
        }
    }
}


# 
# xlate_url()
# ---------
# This function translates the given url so that the pointed document will 
# pass through this CGI for conversion when in CGI mode, or construct a url 
# that fits the needs of the webmaster using the given template, if present. 
# 
sub xlate_url {
    my $url = shift;  ## $url is the url from a href or a src attribute
    my $cgi_options = '';
    
    if($cgi) {
        ## CGI mode
        my $absurl = URI::URL->new($url, $state{doc_uri})->abs;
        
        for my $option (keys %options) {
            $cgi_options .= "$option=$options{$option}\&amp;" if $cgi->param($option)
        }
        
        return $cgi->url . "?${cgi_options}url=$absurl"
        
    } else {
        ## shell mode
        if($options{hreftmpl} and $url !~ m|^http://|) {
            ## we have a template to construct the new url
            my $tmpl = $options{hreftmpl};
            
            my($filename,$filepath,$filetype) = fileparse($url, '((?:\.\w+)+)');
            
            $tmpl =~ s/\{URL\}/$url/;
            $tmpl =~ s/\{FILEPATH\}/$filepath/;
            $tmpl =~ s/\{FILENAME\}/$filename/;
            $tmpl =~ s/\{FILETYPE\}/$filetype/;
            
            return $tmpl
            
        } else {
            return $url
        }
    }
}


# 
# htmlize()
# -------
# This function translate the given text into HTML, and add line numbers. 
# 
sub htmlize {
    my $str = shift;
    my @res = ();
    
    ## convert special chars to entities
    $str =~ s/&/&amp;/go;
    $str =~ s/</&lt;/go;
    $str =~ s/>/&gt;/go;
    
    return "<pre>$str</pre>"
}


# 
# load_entities()
# -------------
# 
sub load_entities {
    %entities = (
        nbsp     => [ 32, ' '], 
        iexcl    => [161, '!'], 
        cent     => [162, '-c-'], 
        pound    => [163, '-L-'], 
        curren   => [164, 'CUR'], 
        yen      => [165, 'YEN'], 
        brvbar   => [166, '|'], 
        sect     => [167, 'S:'], 
        uml      => [168, '"'], 
        copy     => [169, '(C)'], 
        ordf     => [170, '-a'], 
        laquo    => [171, '<<'], 
       'not'     => [172, 'NOT'], 
        shy      => [173, '-'], 
        reg      => [174, '(R)'], 
        macr     => [175, '-'], 
        deg      => [176, 'DEG'], 
        plusmn   => [177, '+/-'], 
        sup2     => [178, '^2'], 
        sup3     => [179, '^3'], 
        acute    => [180, '\''], 
        micro    => [181, 'u'], 
        para     => [182, 'P:'], 
        middot   => [183, '.'], 
        cedil    => [184, ','], 
        sup1     => [185, '^1'], 
        ordm     => [186, '-o'], 
        raquo    => [187, '>>'], 
        frac14   => [188, ' 1/4'], 
        frac12   => [189, ' 1/2'], 
        frac34   => [190, ' 3/4'], 
        iquest   => [191, '?'], 
        Agrave   => [192, 'A'], 
        Aacute   => [193, 'A'], 
        Acirc    => [194, 'A'], 
        Atilde   => [195, 'A'], 
        Auml     => [196, 'Ae'], 
        Aring    => [197, 'A'], 
        AElig    => [198, 'AE'], 
        Ccedil   => [199, 'C'], 
        Egrave   => [200, 'E'], 
        Eacute   => [201, 'E'], 
        Ecirc    => [202, 'E'], 
        Euml     => [203, 'E'], 
        Igrave   => [204, 'I'], 
        Iacute   => [205, 'I'], 
        Icirc    => [206, 'I'], 
        Iuml     => [207, 'I'], 
        ETH      => [208, 'DH'], 
        Ntilde   => [209, 'N'], 
        Ograve   => [210, 'O'], 
        Oacute   => [211, 'O'], 
        Ocirc    => [212, 'O'], 
        Otilde   => [213, 'O'], 
        Ouml     => [214, 'Oe'], 
       'times'   => [215, '*'], 
        Oslash   => [216, 'O'], 
        Ugrave   => [217, 'U'], 
        Uacute   => [218, 'U'], 
        Ucirc    => [219, 'U'], 
        Uuml     => [220, 'Ue'], 
        Yacute   => [221, 'Y'], 
        THORN    => [222, 'P'], 
        szlig    => [223, 'ss'], 
        agrave   => [224, 'a'], 
        aacute   => [225, 'a'], 
        acirc    => [226, 'a'], 
        atilde   => [227, 'a'], 
        auml     => [228, 'ae'], 
        aring    => [229, 'a'], 
        aelig    => [230, 'ae'], 
        ccedil   => [231, 'c'], 
        egrave   => [232, 'e'], 
        eacute   => [233, 'e'], 
        ecirc    => [234, 'e'], 
        euml     => [235, 'e'], 
        igrave   => [236, 'i'], 
        iacute   => [237, 'i'], 
        icirc    => [238, 'i'], 
        iuml     => [239, 'i'], 
        eth      => [240, 'e'], 
        ntilde   => [241, 'n'], 
        ograve   => [242, 'o'], 
        oacute   => [243, 'o'], 
        ocirc    => [244, 'o'], 
        otilde   => [245, 'o'], 
        ouml     => [246, 'o'], 
        divide   => [247, '/'], 
        oslash   => [248, 'o'], 
        ugrave   => [249, 'u'], 
        uacute   => [250, 'u'], 
        ucirc    => [251, 'u'], 
        uuml     => [252, 'u'], 
        yacute   => [253, 'y'], 
        thorn    => [254, 'p'], 
        yuml     => [255, 'y'], 
    );
}


# 
# version()
# -------
sub version {
    print "$program/$version\n"; exit
}


# 
# usage()
# -----
sub usage {
    print <<"USAGE"; exit
usage: $program [options] file

options: 
  --collapse        collapse redundant white spaces and empty paragraphs (default)
  --nocollapse      desactivate the collapse option
  --debug           activate the debug mode
  --hreftmpl=TMPL   set the template for the links reconstruction engine
  --linearize       linearize the tables (default)
  --nolinearize     don't linearize the tables
  --xmlcheck        activate the XML check: output is passed through XML::Parser
  --noxmlcheck      desactivate the XML checker (default)
  
  --help            show this help screen and exit
  --version         show the program name and version and exit
USAGE
}


# 
# cgiusage()
# --------
sub cgiusage {
    print $cgi->header, <<"USAGE"; exit
<html>
<head>
<title>Error</title>
</head>
<body>
<h1>Error</h1>
<p>This CGI was called with incorrect parameters. 
Check your request and try again </p>
</body>
</html>
USAGE
}


